/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "interpreter.h"
#include "string.h"
#include "display.h"

#define __INTERPRETER_C__
#include "interpreter_macro.h"
#include "interpreter_functions.c"

/** Defines the assembler instructions.
 */
static const long ASSEMBLER[] = {
  ASM_NOP,       PASM_NONE, PASM_SDECL("NOP"), PASM_SDECL(NULL),
  ASM_NOOP,      PASM_NONE, PASM_SDECL("NOOP"), PASM_SDECL(NULL),
  PASM_DECL( DECLARE,       PASM_ONE ),
  PASM_DECL( MOVE,          PASM_TWO|PASM_ADR1 ),
  PASM_DECL( MOVA,          PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( LESSER,        PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( GREATER,       PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( EQUAL,         PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( ZERO,          PASM_ONE|PASM_ADR1 ),
  PASM_DECL( JUMP,          PASM_ONE|PASM_ADR1 ),
  PASM_DECL( JUMPIF,        PASM_ONE|PASM_ADR1 ),
  PASM_DECL( JUMPIFNOT,     PASM_ONE|PASM_ADR1 ),
  PASM_DECL( INC,           PASM_ONE|PASM_ADR1 ),
  PASM_DECL( DEC,           PASM_ONE|PASM_ADR1 ),
  PASM_DECL( ADD,           PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( SUB,           PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( MUL,           PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( DIV,           PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( MOD,           PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( RAND,          PASM_ONE|PASM_ADR1 ),
  PASM_DECL( MEMSIZE,       PASM_ONE|PASM_ADR1 ),
  PASM_DECL( CODESEG,       PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( REGSEG,        PASM_TWO|PASM_ADR1|PASM_ADR2 ),
  PASM_DECL( CO,            PASM_ONE|PASM_ADR1 ),
  PASM_DECL( DISPVAL,       PASM_ONE ),
  PASM_DECL( DISP,          PASM_ONE|PASM_ADR1 ),
  PASM_DECL( DISPSTR,       PASM_TWO|PASM_ADR1 ),
  PASM_DECL( END,           PASM_NONE ),
} ;

/** Adds an opponents to the specified opponent's list.
 *
 * @param codes is the list of opponent's codes
 * @param sizes is the quantity of instructions in each opponent's code
 * @param count is the quantity of opponents
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_addOpponent(MEMORY_ITEM*** codes, CODE_SIZE** sizes, OPPONENT_SIZE* count) {
  if ((codes==NULL)&&(sizes==NULL)&&(count==NULL)) return FALSE ;
  
  MEMORY_ITEM** ncodes = (MEMORY_ITEM**)realloc(*codes,((*count)+1)*sizeof(MEMORY_ITEM*));
  CODE_SIZE* nsizes = (CODE_SIZE*)realloc(*sizes,((*count)+1)*sizeof(CODE_SIZE));

  if ((ncodes==NULL)||(nsizes==NULL)) return FALSE ;
  
  ncodes[*count] = NULL ;
  nsizes[*count] = 0 ;
  *codes = ncodes ;
  *sizes = nsizes ;
  (*count)++ ;
  
  return TRUE ;
}

/** Adds an opponent's instruction to the specified opponent's code.
 *
 * @param inst is the identifier of the instruction to tun
 * @param param1 is the first parameter of the instruction or 0
 * @param param2 is the second parameter of the instruction or 0
 * @param codes is the list of opponent's codes
 * @param sizes is the quantity of instructions in each opponent's code
 * @param nopponent is the opponents
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_addOpponentCode(ASSEMBLER_CODE inst, ASSEMBLER_PARAM param1, ASSEMBLER_PARAM param2, MEMORY_ITEM** codes, CODE_SIZE* sizes, OPPONENT_SIZE nopponent) {
  if ((codes==NULL)||(sizes==NULL)) return FALSE ;

  MEMORY_ITEM* ncodes = (MEMORY_ITEM*)realloc(codes[nopponent],
					      sizeof(MEMORY_ITEM)*(sizes[nopponent]+3)) ;
  if (ncodes==NULL) return FALSE ;

  ncodes[sizes[nopponent]] = (MEMORY_ITEM)inst ;
  ncodes[sizes[nopponent]+1] = (MEMORY_ITEM)param1 ;
  ncodes[sizes[nopponent]+2] = (MEMORY_ITEM)param2 ;

  codes[nopponent] = ncodes ;
  sizes[nopponent] += 3 ;

#ifdef __DEBUG__
  d_dbglog("ASSEMBLER: add code (%lu;%lu;%lu)\n",
	   (unsigned long)inst,
	   (unsigned long)param1,
	   (unsigned long)param2);
#endif

  return TRUE ;
}

/** Adds an opponent's data into the specified opponent's code.
 *
 * @param data is the data to add
 * @param codes is the list of opponent's codes
 * @param sizes is the quantity of instructions in each opponent's code
 * @param nopponent is the opponents
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_addOpponentData(ASSEMBLER_PARAM data, MEMORY_ITEM** codes, CODE_SIZE* sizes, OPPONENT_SIZE nopponent) {
  if ((codes==NULL)||(sizes==NULL)) return FALSE ;

  MEMORY_ITEM* ncodes = (MEMORY_ITEM*)realloc(codes[nopponent],
					      sizeof(MEMORY_ITEM)*(sizes[nopponent]+1)) ;
  if (ncodes==NULL) return FALSE ;

  ncodes[sizes[nopponent]] = (MEMORY_ITEM)PASM_MAKEDATASEGMENT(data) ;

  codes[nopponent] = ncodes ;
  sizes[nopponent] ++ ;

#ifdef __DEBUG__
  d_dbglog("ASSEMBLER: add data (%lu)\n", (unsigned long)data) ;
#endif

  return TRUE ;
}

/** Replies the index inside the array of assembler definitions.
 *
 * @param inst is the id of the assembler instruction to find.
 * @param index is the index of the specified instruction.
 * @return 
 */
boolean getAssemblerDef(ASSEMBLER_CODE inst, unsigned long* index) {
  unsigned long i;
  for(i=0; i<ASSEMBLER_SIZE*4; i+=4) {
    if ( ASSEMBLER[i] == inst ) {
      if (index!=NULL) (*index) = i ;
      return TRUE ;
    }
  }
  return FALSE ;
}

/** Display an instruction parameter.
 *
 * @param param is the parameter to display
 * @param flow is the flow in which the print must be done
 */
void displayInstructionParam(ASSEMBLER_PARAM param, FILE* flow) {
  if ( PASM_ISADRREDIRECTION(param) ) {
    param = PASM_UNMAKEADRREDIRECTION(param) ;
    if (PASM_ISREGISTER(param)) {
      display(flow," (R%lu)", (unsigned long)PASM_UNMAKEREGISTER(param)) ;
    }
    else if (PASM_ISABSOLUTEADDRESS(param)) {
      display(flow," (#%lu)", (unsigned long)PASM_UNMAKEABSOLUTEADDRESS(param)) ;
    }
    else {
      display(flow," (%lu)", (unsigned long)param) ;
    }
  }
  else {
    if (PASM_ISREGISTER(param)) {
      display(flow," R%lu", (unsigned long)PASM_UNMAKEREGISTER(param)) ;
    }
    else if (PASM_ISABSOLUTEADDRESS(param)) {
      display(flow," #%lu", (unsigned long)PASM_UNMAKEABSOLUTEADDRESS(param)) ;
    }
    else {
      display(flow," %lu", (unsigned long)param) ;
    }
  }
}

/** Display an instruction.
 *
 * @param inst is the instruction to display
 * @param param1 is the first parameter
 * @param param2 is the second parameter
 * @param flow is the flow in which the print must be done
 */
void displayInstruction(ASSEMBLER_CODE inst, ASSEMBLER_PARAM param1, ASSEMBLER_PARAM param2,FILE* flow) {
  unsigned long index ;
  if ( getAssemblerDef(inst,&index) ) {
    unsigned long proto = ASSEMBLER[index+1] ;
    display(flow,"%s", (char*)(ASSEMBLER[index+2])) ;
    if (proto&(PASM_TWO|PASM_ONE)) {
      displayInstructionParam(param1,flow) ;
    }
    if (proto&PASM_TWO) {
      displayInstructionParam(param2,flow) ;
    }
  }
  else {
    display(flow,"?") ;
  }
}

/** Display the codes of the opponents.
 *
 * @param codes is the list of opponent's codes
 * @param sizes is the quantity of instructions in each opponent's code
 * @param count is the quantity of opponents
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
void ASM_displayCodes(MEMORY_ITEM** codes, CODE_SIZE* sizes, OPPONENT_SIZE count) {
  unsigned long i,j;
  d_log("Source codes for %u opponents:\n",count) ;
  for(i=0; i<count; i++) {
    d_log("\tOpponent %u (%u bytes):\n", i,sizes[i]) ;
    j=0 ;
    while (j<sizes[i]) {
      if ( PASM_ISDATASEGMENT(codes[i][j]) ) {
	/* data segment */
	d_log("\t|\t%u:\t0x%X\t",j,(short)codes[i][j]);
	if ( isprint((char)codes[i][j]) ) {
	  d_log("'%c'",(char)codes[i][j]) ;
	}
	d_log("\t\tDATA\n") ;
	j++ ;
      }
      else if ((j+2)<sizes[i]) {
	/* code segment */
	d_log("\t|\t%u:\t0x%X\t0x%X\t0x%X\t",j,(short)codes[i][j],(short)codes[i][j+1],(short)codes[i][j+2]) ;
	displayInstruction(codes[i][j],codes[i][j+1],codes[i][j+2],getLogFlow()) ;
	d_log("\n") ;
	j += 3 ;
      }
      else {
	/* something wrong */
	for(;j<sizes[i];j++) {
	  d_log("\t|\t%u:\t0x%X\t",j,(short)codes[i][j]);
	  if ( isprint((char)codes[i][j]) ) {
	    d_log("'%c'",(char)codes[i][j]) ;
	  }
	  d_log("\t\tDATA\n") ;
	}
      }
    }
  }
}

/** Run the specified assembler instruction.
 *
 * @param vm is the virtual machine inside which the instruction
 *        must be ran
 * @param inst is the identifier of the instruction to tun
 * @param param1 is the first parameter of the instruction or 0
 * @param param2 is the second parameter of the instruction or 0
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_run(VIRTUAL_MACHINE* vm, ASSEMBLER_CODE inst, ASSEMBLER_PARAM param1, ASSEMBLER_PARAM param2) {
  if (vm!=NULL) {

    /* Get the instruction definition */
    unsigned long index ;
    if ( getAssemblerDef(inst,&index) ) {
      unsigned long proto = ASSEMBLER[index+1] ;
      void* fct = (void*)ASSEMBLER[index+3] ;
     
#ifdef __DEBUG__
      d_dbglog("ASSEMBLER: run(%lu@%lu) = ",
	       (unsigned long)VM_getCurrentPlayer(vm),
	       (unsigned long)VM_getOrdinalCounter(vm)) ;
      displayInstruction(inst,param1,param2,getDbgLogFlow()) ;
      d_dbglog("\n") ;
#endif

      if ( fct != NULL ) {
	/* run the instruction's function accordingly
	   to the quantity of parameters */
	if (proto&PASM_TWO) {
 	  boolean (*f)(VIRTUAL_MACHINE* m, ASSEMBLER_PARAM p1, ASSEMBLER_PARAM p2) = fct ;
	  return f(vm,param1,param2) ;
	}
	else if (proto&PASM_ONE) {
	  boolean (*f)(VIRTUAL_MACHINE* m, ASSEMBLER_PARAM p1) = fct ;
	  return f(vm,param1) ;
	}
	else {
	  boolean (*f)(VIRTUAL_MACHINE* m) = fct ;
	  return f(vm) ;
	}
      }
      else {
	return !(proto&PASM_NONE) ;
      }

    }

  }
  return FALSE ;
}

/** Translates the addresses of the specified instruction.
 * It is needed because the addreses of a program are
 * always given from the start of the program code.
 * This translation permits to update the addresses to
 * be compliant with the real location of the source code.
 *
 * @param inst is the identifier of the instruction to tun
 * @param param1 is the first parameter of the instruction or 0
 * @param param2 is the second parameter of the instruction or 0
 * @param vm is the virtual machine inside which the instruction
 *        must be ran
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_translateAddresses(ASSEMBLER_CODE inst, ASSEMBLER_PARAM* param1, ASSEMBLER_PARAM* param2, MEMORY_ADDRESS base) {
  unsigned long index ;
  if ( getAssemblerDef(inst,&index) ) {
    unsigned long proto = ASSEMBLER[index+1] ;
    if (proto&PASM_ADR1) {
      if (param1==NULL) return FALSE ;
      if ((!PASM_ISREGISTER(*param1))&&
	  (!PASM_ISABSOLUTEADDRESS(*param1))) (*param1) += base ;
    }
    if (proto&PASM_ADR2) {
      if (param2==NULL) return FALSE ;
      if ((!PASM_ISREGISTER(*param2))&&
	  (!PASM_ISABSOLUTEADDRESS(*param2))) (*param2) += base ;
    }
    return TRUE ;
  }
  return FALSE ;
}

/** Replies the identifier that corresponds to the
 *  assembler instruction which is passed in parameter.
 *
 * @param token is the string which corresponds to the instruction to search
 * @param inst is the identifier of the instruction
 * @param proto is the prototype of the found instruction
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean ASM_getInstructionDef(char* token, ASSEMBLER_CODE* inst, unsigned char* proto) {
  unsigned long i,len;
  if ((token==NULL)||
      (inst==NULL)||
      (proto==NULL)) return FALSE ;

  len = strlen(token) ;
  for(i=0; i<ASSEMBLER_SIZE*4; i+=4) {
    if (((char*)ASSEMBLER[i+2])!=NULL) {
      unsigned long len2 = strlen((char*)ASSEMBLER[i+2]) ;
      len2 = (len>len2)?len:len2 ;
      if ( compareStrings( (char*)ASSEMBLER[i+2],
			   token, len2 ) == 0 ) {
	*inst = ASSEMBLER[i] ;
	*proto = (unsigned char)ASSEMBLER[i+1] ;
	return TRUE ;
      }
    }
  }

  return FALSE ;
}

/** Replies the prototype of the specified assembler function.
 *
 * @param inst is the identifier of the instruction
 * @return the prototype of 0
 */
unsigned char ASM_getInstructionPrototype(ASSEMBLER_CODE inst) {
  unsigned long i ;
  for(i=0; i<ASSEMBLER_SIZE*4; i+=4) {
    if (ASSEMBLER[i]==inst) {
      return (unsigned char)ASSEMBLER[i+1] ;
    }
  }
  return 0 ;
}
