/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/
#ifndef __EW_MEMORY_H__
#define __EW_MEMORY_H__

#include "boolean.h"
#include "types.h"

/** Defines the NULL pointer for the virtual memory.
 */
#define VNULL ((MEMORY_ADDRESS)-1)

/** Data structure for a virtual memory.
 *  It includes:
 *  - the content of the memory (an set of
 *    bytes)
 *  - the size of the memory (in bytes)
 */
typedef struct {
  MEMORY_ITEM* content ;
  MEMORY_SIZE size ;
} VIRTUAL_MEMORY ;

/** Initializes the specified virtual memory.
 *  This function must always be called before
 *  any operation on the memory.
 *
 * @param memory is the memory to initialize
 * @param size is the desired size (in bytes) of the memory
 */
void MEM_init(VIRTUAL_MEMORY* memory, MEMORY_SIZE size) ;

/** Destroyes the specified virtual memory.
 *  This function must be called after all the
 *  operations on the specified memory.
 *
 * @param memory is the memory to delete
 */
void MEM_delete(VIRTUAL_MEMORY* memory) ;

/** Replies if the specified memory address is
 *  valid for the specified memory.
 *  An address is valid only if it's corresponding
 *  to an item (data) of the memory.
 *
 * @param memory is the memory to use
 * @param adr is the address to check
 * @return TRUE if the adr is the address of a memory element,
 *         otherwhise FALSE
 */
boolean MEM_isValidAdr(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr) ;

/** Replies if the specified memory address is
 *  valid for an instruction storage.
 *
 * @param memory is the memory to use
 * @param count is the quantity of data to set
 * @return TRUE if the adr is the address of a memory element,
 *         otherwhise FALSE
 */
boolean MEM_isValidDataAdr(VIRTUAL_MEMORY* memory, MEMORY_SIZE count, ...) ;

/** Replies if the specified set of contigue memory addresses is
 *  valid.
 *
 * @param memory is the memory to use
 * @param adr is the address to check
 * @param count is the quantity of addresses
 * @return TRUE if the adr is the address of a memory element,
 *         otherwhise FALSE
 */
boolean MEM_isValidContigueAdr(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr, MEMORY_SIZE count) ;

/** Replies if the size in bytes of the memory.
 *
 * @param memory is the memory to use
 * @return the size in bytes of the memory
 */
MEMORY_SIZE MEM_size(VIRTUAL_MEMORY* memory) ;

/** Put a data at the specified address of the memory.
 *
 * @param memory is the memory to use
 * @param adr is the address at which the data must be put
 * @param data is the data to put
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean MEM_set(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr, MEMORY_ITEM value) ;

/** Replies the data at the specified address of the memory.
 *
 * @param memory is the memory to use.
 * @param adr is the address at which the data must be put.
 * @return the data at the specified address or 0 if
 *         the specified address was wrong.
 */
MEMORY_ITEM MEM_get(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr) ;

/** Put a set of data at the specified address of the memory.
 *
 * @param memory is the memory to use
 * @param adr is the address at which the data must be put
 * @param count is the quantity of data to set
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean MEM_setData(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr, MEMORY_SIZE count, ...) ;

/** Replies the set of data at the specified address of the memory.
 *
 * @param memory is the memory to use
 * @param adr is the address at which the data must be put
 * @param count is the quantity of data to set
 * @return TRUE if something was replied, otherwhise FALSE
 */
boolean MEM_getData(VIRTUAL_MEMORY* memory, MEMORY_ADDRESS adr, MEMORY_SIZE count, ...) ;

#endif
