/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/

#include <stdlib.h>
#include <time.h>

#include "vm.h"
#include "interpreter.h"
#include "display.h"

#define __VM_C__
#include "vm_private.c"

/** Initializes the virtual machine.
 *
 * @param vm is the virtual machine to initialize.
 * @param nopponents is the quantity of opponents supported by the virtual machine
 * @param memsize is the size in bytes of the virtual memory
 */
void VM_init(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponents, MEMORY_SIZE memsize) {
  unsigned long i ;
  if ((vm!=NULL)&&(nopponents>0)) {
    unsigned long limit = THEORICAL_MEMORY_LIMIT ;

    vm->nopponents = nopponents ;
    vm->current_player = 0 ;
    vm->turns = 0 ;

    /* memory allocation */
    vm->ordinal_counters = (MEMORY_ADDRESS*)malloc(sizeof(MEMORY_ADDRESS)*nopponents) ;
    vm->boolean_tests = (boolean*)malloc(sizeof(boolean)*nopponents) ;
    vm->error_messages = (ERRORNUM*)malloc(sizeof(ERRORNUM)*nopponents) ;
    vm->array_decls = (MEMORY_ADDRESS*)malloc(sizeof(MEMORY_ADDRESS)*nopponents*2) ;
    vm->code_segments = (MEMORY_ADDRESS*)malloc(sizeof(MEMORY_ADDRESS*)*nopponents*2) ;
    MEM_init(&(vm->memory),memsize);

    /* initialization */
    for(i=0; i<nopponents; i++) {
      vm->ordinal_counters[i] = 0 ;
      vm->boolean_tests[i] = NO_BOOL_TEST ;
      vm->error_messages[i] = ERR_NONE ;
      vm->array_decls[i*2] = NO_ARRAY_DECL ;
      vm->array_decls[i*2+1] = NO_ARRAY_DECL ;
      vm->code_segments[i*2] = VNULL ;
      vm->code_segments[i*2+1] =  VNULL ;
    }

    srand(time(NULL));

    d_log("\telements per instruction: %lu\n", (unsigned long)ASM_INSTRUCTION_SIZE) ;
    d_log("\tinstruction size: %lu byte(s)\n", (unsigned long)(ASM_INSTRUCTION_SIZE*sizeof(MEMORY_ITEM))) ;
    d_log("\ttheorical memory limits: [0;%lu]\n", limit) ;
    d_log("\tallocated memory inside the theorical limits? %s\n", 
	  (MEM_size(&(vm->memory))<=(limit+1)) ? "Yes" : "No" ) ;
    d_log("\topponents: %lu\n", (unsigned long)nopponents) ;
    d_log("\tturn limit: %lu\n", (unsigned long)VM_TURN_MAX) ;
  }
}

/** Destroyes the virtual machine.
 *
 * @param vm is the virtual machine to delete.
 */
void VM_delete(VIRTUAL_MACHINE* vm) {
  if (vm!=NULL) {
    MEM_delete(&(vm->memory));
    if (vm->boolean_tests!=NULL) free(vm->boolean_tests) ;
    if (vm->ordinal_counters!=NULL) free(vm->ordinal_counters) ;
    if (vm->error_messages!=NULL) free(vm->error_messages) ;
    if (vm->array_decls!=NULL) free(vm->array_decls) ;
    if (vm->code_segments!=NULL) free(vm->code_segments) ;
    vm->nopponents = 0 ;
    vm->current_player = 0 ;
    vm->turns = 0 ;
    vm->boolean_tests = NULL ;
    vm->ordinal_counters = NULL ;
    vm->error_messages = NULL ;
    vm->array_decls = NULL ;
    vm->code_segments = NULL ;
  }
}

/** Loads the code of an opponent.
 *  This operation erase the previously loaded opponent
 *  at the same index.
 *  This method does not check if any other opponent has
 *  any source code where the loaded opponent was put.
 *
 * @param vm is the virtual machine to use.
 * @param nopponent is the index of the opponent to load.
 * @param code is the source code to load into the memory.
 * @param size is the source code's size.
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_unsecureOpponentLoading(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponent, MEMORY_ITEM* code, CODE_SIZE size) {
  MEMORY_ADDRESS start = (MEMORY_ADDRESS)(rand() * 
					  (MEM_size(&(vm->memory)-size))
					  / RAND_MAX) ; ;
  return unsecureOpponentLoading(vm,start,nopponent,code,size) ;
}

/** Loads the code of all the opponents.
 *  This operation erase the previously loaded opponents.
 *  This method ensure all the proposed source code will not
 *  be put in same memory items.
 *
 * @param vm is the virtual machine to use.
 * @param nopponents is the quantity of opponents to load
 * @param codes is the array of source codes.
 * @param sizes is the array of source code's sizes.
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_secureOpponentLoading(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponents, MEMORY_ITEM** codes, CODE_SIZE* sizes) {
  MEMORY_ADDRESS* segments = NULL ;
  unsigned short i ;
  unsigned char retry = 0 ;
  boolean intersection = TRUE ;

  while ((retry<MAX_RETRY)&&(intersection)) {
    /* Create the allocation segments */
    dispatchSourceMemory(vm,nopponents,&segments,sizes) ;
    retry ++ ;
    /* try to find any intersect */
    intersection = findSourceIntersection(nopponents,segments) ;
  }

  if (intersection) {
    /* disallocate memory */
    if ( segments != NULL ) {
      free(segments) ;
    }
    return FALSE ;
  }

  /* found one allocation */
  for(i=0; i<nopponents; i++) {
    if ( ! unsecureOpponentLoading(vm,
				   segments[i*2], 
				   i, codes[i],
				   sizes[i]) ) {
      /* disallocate memory */
      if ( segments != NULL ) {
	free(segments) ;
      }
      return FALSE ;
    }
  }

  /* disallocate memory */
  if ( segments != NULL ) {
    free(segments) ;
  }
  return TRUE ;
}

/** Changes the ordinal counter for the current opponent.
 *
 * @param vm is the virtual machine to use.
 * @param counter is the new value of the ordinal counter.
 */
void VM_setOrdinalCounter(VIRTUAL_MACHINE* vm, MEMORY_ADDRESS counter) {
  if (vm!=NULL) {
    vm->ordinal_counters[vm->current_player] = counter ;
  }
}

/** Replies the ordinal counter for the current opponent.
 *
 * @param vm is the virtual machine to use.
 * @return an ordinal counter
 */
MEMORY_ADDRESS VM_getOrdinalCounter(VIRTUAL_MACHINE* vm) {
  if (vm!=NULL) {
    return vm->ordinal_counters[vm->current_player] ;
  }
  return 0 ;
}

/** Replies the ordinal counter for the specified opponent.
 *
 * @param vm is the virtual machine to use.
 * @param nopponent is the index of the opponent
 * @return an ordinal counter
 */
MEMORY_ADDRESS VM_getOrdinalCounterFor(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponent) {
  if ((vm!=NULL)&&(nopponent<vm->nopponents)) {
    return vm->ordinal_counters[nopponent] ;
  }
  return 0 ;
}

/** Replies the current playing opponent.
 *
 * @param vm is the virtual machine to use.
 * @return the current playing opponent
 */
OPPONENT_SIZE VM_getCurrentPlayer(VIRTUAL_MACHINE* vm) {
  if (vm!=NULL) {
    return vm->current_player ;
  }
  return 0 ;
}

/** Replies the result of the previously saved test.
 *
 * @param vm is the virtual machine to use.
 * @param result is the result
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_getLastTest(VIRTUAL_MACHINE* vm, boolean* result) {
  if ((vm!=NULL)&&
      (result!=NULL)&&
      (vm->boolean_tests[vm->current_player]!=NO_BOOL_TEST)) {
    *result = vm->boolean_tests[vm->current_player] ;
    return TRUE ;
  }
  return FALSE ;
}

/** Set the result of the last test.
 *
 * @param vm is the virtual machine to use.
 * @param result is the result
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_setLastTest(VIRTUAL_MACHINE* vm, boolean result) {
  if (vm!=NULL) {
    vm->boolean_tests[vm->current_player] = result ;
    return TRUE ;
  }
  return FALSE ;
}

/** Replies the base address of the last array declaration.
 *
 * @param vm is the virtual machine to use.
 * @param base is the base address
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_getArrayBaseAddress(VIRTUAL_MACHINE* vm, MEMORY_ADDRESS* base) {
  if ((vm!=NULL)&&
      (base!=NULL)&&
      (vm->array_decls[vm->current_player*2]!=NO_ARRAY_DECL)) {
    *base = vm->array_decls[vm->current_player*2] ;
    return TRUE ;
  }
  return FALSE ;  
}

/** Replies the addresses of the last array declaration.
 *
 * @param vm is the virtual machine to use.
 * @param base is the base address
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_getArrayAddresses(VIRTUAL_MACHINE* vm, MEMORY_ADDRESS* baseadr, MEMORY_ADDRESS* endadr) {
  if ((vm==NULL)||
      ((baseadr!=NULL)&&
       (vm->array_decls[vm->current_player*2]==VNULL))||
      ((endadr!=NULL)&&
       (vm->array_decls[vm->current_player*2+1]==VNULL)))
    return FALSE ;
  if (baseadr!=NULL) *baseadr = vm->array_decls[vm->current_player*2] ;
  if (endadr!=NULL) *endadr = vm->array_decls[vm->current_player*2+1] ;
  return TRUE ;
}

/** Replies the addresses of the code segment.
 *
 * @param vm is the virtual machine to use.
 * @param baseadr is the base address
 * @param endadr is the end address
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_getCodeSegmentAddresses(VIRTUAL_MACHINE* vm, MEMORY_ADDRESS* baseadr, MEMORY_ADDRESS* endadr) {
  if ((vm==NULL)||
      ((baseadr!=NULL)&&
       (vm->array_decls[vm->current_player*2]==NO_ARRAY_DECL))||
      ((endadr!=NULL)&&
       (vm->array_decls[vm->current_player*2+1]==NO_ARRAY_DECL)))
    return FALSE ;
  if (baseadr!=NULL) *baseadr = vm->code_segments[vm->current_player*2] ;
  if (endadr!=NULL) *endadr = vm->code_segments[vm->current_player*2+1] ;
  return TRUE ;
}


/** Sets the base address of the last array declaration.
 *
 * @param vm is the virtual machine to use.
 * @param length is the length of the array
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_computeArrayBaseAddress(VIRTUAL_MACHINE* vm, MEMORY_SIZE length) {
  if ((vm!=NULL)&&
      (length>0)) {
    unsigned long retry=0 ;
    boolean intersection=TRUE ;
    MEMORY_ADDRESS segment[2] ;
    
    while ((retry<MAX_RETRY)&&(intersection)) {
      /* Create the array */
      dispatchMemorySegment(vm,length,segment) ;
      retry ++ ;
      /* try to find any intersect */
      intersection = findMemoryIntersection(vm,vm->current_player,segment) ;
    }

    
    if (intersection) {
      /* disallocate memory */
      if ( segment != NULL ) {
	free(segment) ;
      }
      VMERROR(ERR_MEM_NOTENOUGHCONTIGUE);
      return FALSE ;
    }    
    
    /* found a valid segment, register it */    
    vm->array_decls[vm->current_player*2] = segment[0] ;
    vm->array_decls[vm->current_player*2+1] = segment[1] ;

    return TRUE ;
  }
  return FALSE ;  
}

/** Replies the address to the virtual memory.
 *
 * @param vm is the virtual machine to use
 * @return the address of the current virtual memory
 */
VIRTUAL_MEMORY* VM_getVirtualMemory(VIRTUAL_MACHINE* vm) {
  if (vm!=NULL) {
    return &(vm->memory) ;
  }
  return NULL ;
}

/** Replies the winner of the last play.
 *
 * @param vm is the virtual machine to use.
 * @return the index of the winner or -1.
 */
OPPONENT_SIZE VM_getWinner(VIRTUAL_MACHINE* vm) {
  if (vm!=NULL) {
    OPPONENT_SIZE i,count,idx ;
    for(i=0,count=0; (i<vm->nopponents)&&(count<2); i++) {
      if (vm->ordinal_counters[i]!=VNULL) {
	count++;
	idx = i ;
      }
    }
    if (count==1) {
      return idx ;
    }
  }
  return (OPPONENT_SIZE)-1;
}

/** Sets the error message for the current opponent.
 * The message will NOT be freed by the destructor of
 * the virtual machine.
 *
 * @param vm is the virtual machine to delete.
 * @param message is the error message.
 */
void VM_setErrorMsg(VIRTUAL_MACHINE* vm, ERRORNUM message) {
  if (vm!=NULL) {
    vm->error_messages[vm->current_player] = message ;
  }
}

/** Replies the error number for the specified opponent.
 *
 * @param vm is the virtual machine to delete.
 * @param n is the index of the opponent
 * @return the error number.
 */
ERRORNUM VM_getError(VIRTUAL_MACHINE* vm, OPPONENT_SIZE opponent) {
  if ((vm!=NULL)&&(opponent<vm->nopponents)) {
    return vm->error_messages[opponent] ;
  }
  return ERR_NONE ;
}

/** Replies the error message for the specified opponent.
 *
 * @param vm is the virtual machine to delete.
 * @param n is the index of the opponent
 * @return the error message.
 */
const char* VM_getErrorMsg(VIRTUAL_MACHINE* vm, OPPONENT_SIZE opponent) {
  ERRORNUM err = VM_getError(vm,opponent) ;
  return error_msg(err) ;
}

/** Stop the current opponent.
 *
 * @param vm is the virtual machine to delete.
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean VM_stopOpponent(VIRTUAL_MACHINE* vm) {
  if (vm!=NULL) {
    vm->ordinal_counters[vm->current_player] = VNULL-ASM_INSTRUCTION_SIZE ;
    return TRUE ;
  }
  return FALSE ;  
}

/** Replies the quantity of turns in the last game.
 *
 * @return the quantity of turns.
 */
unsigned long long VM_getTurns(VIRTUAL_MACHINE* vm) {
  return (vm!=NULL) ? vm->turns : 0 ;
}

/** Run the virtual machine.
 *
 * @param vm is the virtual machine to use.
 * @return TRUE if an opponent win, otherwhise FALSE
 */
boolean VM_run(VIRTUAL_MACHINE* vm) {
  if (vm!=NULL){
    OPPONENT_SIZE* playingorder=NULL ;
    OPPONENT_SIZE player ;
    
    /* Initialize the playing order array */
    playingorder = (OPPONENT_SIZE*)malloc(sizeof(OPPONENT_SIZE)*vm->nopponents) ;
#ifdef __DEBUG__
    d_log("Ordinal counter are:\n") ;
#endif
    for(player=0; player<vm->nopponents; player++) {
      playingorder[player] = player ;
#ifdef __DEBUG__
      d_log("\topponent(%lu) = %lu\n",
	    (unsigned long)player,
	    (unsigned long)VM_getOrdinalCounterFor(vm,player)) ;      
#endif
    }
    
    vm->turns = 0 ;
    
    while ((vm->turns<VM_TURN_MAX)&&(hasOpponents(vm))) {

      vm->turns++ ;
      d_log("starting turn %lu\n",vm->turns);

      /* compute a new playing order */
      shufflePlayers(playingorder,vm->nopponents) ;

      /* Run a turn */
      for(player=0;player<vm->nopponents;player++) {
	MEMORY_ADDRESS co ;
	
	/* the the current player */
	vm->current_player = playingorder[player] ;

	/* get ordinal counter and check its validity */
	co = VM_getOrdinalCounter(vm) ;
	
	if ( co > 0 ) {

	  /* ordinal counter indicates that an
	     instruction must be ran */

	  if (MEM_isValidContigueAdr(&(vm->memory),co,
				     ASM_INSTRUCTION_SIZE)) {
	    /* ordinal counter is good */

	    /* get the instruction */
	    MEMORY_ITEM instruction, param1, param2 ;
	    if (MEM_getData(&(vm->memory),co,3,&instruction,&param1,&param2)) {
	      /* run the instruction */
	      if ( ASM_run(vm,
			   (ASSEMBLER_CODE)instruction, 
			   (ASSEMBLER_PARAM)param1,
			   (ASSEMBLER_PARAM)param2) ) {
		/* pass to the next instruction */
		co = VM_getOrdinalCounter(vm) + ASM_INSTRUCTION_SIZE ;
	      }
	      else {
		/* something wrong append during the instruction running */
		co = VNULL ;
		if ( VM_getError(vm,vm->current_player) == ERR_NONE ) {
		  VMERROR(ERR_ASM_INSTRUCTIONFAILED) ;
		}
	      }
	    }
	    else {
	      /* something wrong append */
	      co = VNULL ;
	      VMERROR(ERR_VM_INVALIDORDINALCOUNTER) ;
	    }
	  }
	  else {
	    /* ordinal counter corresponds to an
	       invalid memory address */
	    co = VNULL ;
	    VMERROR(ERR_VM_INVALIDORDINALCOUNTER) ;
	  }
	  
	  /* Save the ordinal counter and
	     pass to the next opponent */
	  VM_setOrdinalCounter(vm,co) ;

	}

      }

      vm->current_player = 0 ;

    }

    if ( playingorder!=NULL ) {
      free(playingorder) ;
    }
    
    /* End of the game */
    return ( hasWinner(vm) ) ;

  }
  return FALSE ;
}
