/* Copyright (C) 2004 Stephane GALLAND <stephane.galland@utbm.fr>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.
                                                                               
   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
   This program is free software; you can redistribute it and/or modify
*/

#ifndef __VM_C__
#error Cant compile vm_private.c outside vm.c
#endif

/** Defines the value which indicates that
 *  no boolean test was previously did.
 */
#define NO_BOOL_TEST ((boolean)-1)

/** Defines the value which indicates that
 *  no array definition was previously did.
 */
#define NO_ARRAY_DECL ((MEMORY_ADDRESS)0)

#define MAX_RETRY 100

#define ISBETWEEN(__a__,__b__,__c__) ((__a__>=__b__)&&(__a__<=__c__))
#define ISINSIDE(__a__,__b__,__c__,__d__) (ISBETWEEN(__a__,__c__,__d__)||ISBETWEEN(__b__,__c__,__d__))
#define HASINTERSECT(__a__,__b__,__c__,__d__) (ISINSIDE(__a__,__b__,__c__,__d__)||ISINSIDE(__c__,__d__,__a__,__b__))

/** Replies if the virtual machine has opponents
 *  which are able to fight.
 *
 * @param vm is the virtual machine to use
 * @return TRUE if two or more opponents are always alive,
 *         otherwhise FALSE
 */
boolean hasOpponents(VIRTUAL_MACHINE* vm) {
  unsigned short i,count ;
  for(i=0,count=0; (i<vm->nopponents)&&(count<2); i++) {
    if (vm->ordinal_counters[i]!=VNULL) {
      count++;
    }
  }
  return (count>1) ;
}

/** Replies if the virtual machine has winning opponents.
 * An opponent winning if it was the last leaving opponent.
 *
 * @param vm is the virtual machine to use
 * @return TRUE if two or more opponents are always alive,
 *         otherwhise FALSE
 */
boolean hasWinner(VIRTUAL_MACHINE* vm) {
  OPPONENT_SIZE i,count ;
  for(i=0,count=0; (i<vm->nopponents)&&(count<2); i++) {
    if (vm->ordinal_counters[i]!=VNULL) {
      count++;
    }
  }
  return (count==1) ;
}


/** Loads the code of an opponent.
 *  This operation erase the previously loaded opponent
 *  at the same index.
 *  This method does not check if any other opponent has
 *  any source code where the loaded opponent was put.
 *
 * @param vm is the virtual machine to use.
 * @param beginAt is the address of at which the code must be put
 * @param nopponent is the index of the opponent to load.
 * @param code is the source code to load into the memory.
 * @param size is the source code's size.
 * @return FALSE if something wrong occurs, otherwhise TRUE
 */
boolean unsecureOpponentLoading(VIRTUAL_MACHINE* vm,
				MEMORY_ADDRESS beginAt,
				OPPONENT_SIZE nopponent, MEMORY_ITEM* code,
				CODE_SIZE size) {
  if (vm!=NULL) {
    CODE_SIZE i=0 ;
    MEMORY_ADDRESS adr = beginAt ;
    while (i<size) {

      if ( PASM_ISDATASEGMENT(code[i]) ) {

	/* data segment */

	if ( ! MEM_isValidAdr(&(vm->memory),adr) ) return FALSE ;
	MEM_set(&(vm->memory), adr, PASM_UNMAKEDATASEGMENT(code[i]) ) ;
	i++ ;
	adr ++ ;
      }
      else if ((i+2)<size) {

	/* code segment */

	if (!MEM_isValidContigueAdr(&(vm->memory),adr,
				    ASM_INSTRUCTION_SIZE))
	  return FALSE ;

	ASM_translateAddresses(code[i], code+i+1, code+i+2,beginAt) ;
	if ( code[i] != ASM_NOOP ) {
	  MEM_setData(&(vm->memory), adr,
		      ASM_INSTRUCTION_SIZE, 
		      code[i], code[i+1], code[i+2] ) ;
	  adr += ASM_INSTRUCTION_SIZE ;
	  i += 3 ;
	}

      }
      else {
	/* something wrong */
	for(;i<size;i++,adr++) {
	  if ( ! MEM_isValidAdr(&(vm->memory),adr) ) return FALSE ;
	  MEM_set(&(vm->memory), adr, 
		  (PASM_ISDATASEGMENT(code[i]) ?
		   PASM_UNMAKEDATASEGMENT(code[i]) : code[i] ) ) ;
	}
      }
    }
    vm->ordinal_counters[nopponent] = beginAt ;
    vm->boolean_tests[nopponent] = NO_BOOL_TEST ;
    vm->array_decls[nopponent*2] = NO_ARRAY_DECL ;
    vm->array_decls[nopponent*2+1] = NO_ARRAY_DECL ;
    vm->code_segments[nopponent*2] = beginAt ;
    vm->code_segments[nopponent*2+1] = beginAt+size ;
    return TRUE ;
  }    
  return FALSE ;
}

/** Found an allocation of the source code.
 *
 * @param nopponents is the quantity of opponents
 * @param segments is the memory allocation for source code
 * @param sizes is the array of source code's sizes.
 * @param vm is the virtual machine to use
 */
void dispatchSourceMemory(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponents, MEMORY_ADDRESS** segments, CODE_SIZE* sizes) {
  OPPONENT_SIZE i ;
  *segments = (MEMORY_ADDRESS*)malloc(nopponents*2*sizeof(MEMORY_ADDRESS)) ;

  for(i=0; i<nopponents; i++) {
    int r = rand() ;
    MEMORY_SIZE windowsize = MEM_size(&(vm->memory))-sizes[i] ;
    (*segments)[i*2] = (MEMORY_ADDRESS)((float)r * ((float)windowsize/(float)RAND_MAX)) ;
    (*segments)[i*2+1] = (MEMORY_ADDRESS)((*segments)[i*2] + sizes[i] - 1) ;
  }
}

/** Replies if some code segment has intersection.
 *
 * @param nopponents is the quantity of opponents
 * @param segments is the code segments
 * @return TRUE if two code segments have intersection,
 *         otherwhise FALSE
 */
boolean findSourceIntersection(OPPONENT_SIZE nopponents,MEMORY_ADDRESS* segments) {
  if (segments!=NULL) {
    OPPONENT_SIZE i,j;

    for(i=0; i<nopponents-1; i++) {
      for(j=i+1; j<nopponents; j++) {
	if ( HASINTERSECT(segments[i*2],segments[i*2+1],
			  segments[j*2],segments[j*2+1]) ) {
	  return TRUE ;
	}
      }
    }
  }
  return FALSE ;
}

/** Found an allocation of the memory segment.
 *
 * @param vm is the virtual machine to use
 * @param size is the size to allocate
 * @param segments is the memory allocation
 */
void dispatchMemorySegment(VIRTUAL_MACHINE* vm, MEMORY_SIZE size, MEMORY_ADDRESS segment[2]) {
  int r;
  MEMORY_SIZE windowsize ;
  r = rand() ;
  windowsize = MEM_size(&(vm->memory))-size ;
  do {
    segment[0] = (MEMORY_ADDRESS)((float)r * ((float)windowsize/(float)RAND_MAX)) ;
  }
  while ( segment[0] == NO_ARRAY_DECL ) ;
  segment[1] = (MEMORY_ADDRESS)(segment[0] + size - 1) ;
}

/** Replies if some memory segment has intersection with
 *  another memory segment.
 *
 * @param vm is the virtual machine to use
 * @param nopponent is the index of the opponent to ignore
 * @param segment is the segment to test (array of 2 elements)
 * @return TRUE if two code segments have intersection,
 *         otherwhise FALSE
 */
boolean findMemoryIntersection(VIRTUAL_MACHINE* vm, OPPONENT_SIZE nopponent, MEMORY_ADDRESS segment[2]) {
  if (segment!=NULL) {
    OPPONENT_SIZE i;

    for(i=0; i<vm->nopponents; i++) {

      /* check with another array definitions */
      if ((i!=nopponent)&&
	  (vm->array_decls[i*2]!=NO_ARRAY_DECL)&&
	  ( HASINTERSECT(segment[0],segment[1],
			 vm->array_decls[i*2],
			 vm->array_decls[i*2+1]) )) {
	return TRUE ;
      }

      /* check with code segments */
      if (HASINTERSECT(segment[0],segment[1],
		       vm->code_segments[i*2],
		       vm->code_segments[i*2+1])) {
	return TRUE ;
      }

    }
  }
  return FALSE ;
}

/** Shuffles the specified array of players.
 *
 * @param array is the array to shuffle
 * @param count is the size of the array
 */
void shufflePlayers(OPPONENT_SIZE* array, OPPONENT_SIZE count) {
  OPPONENT_SIZE n ;
  const OPPONENT_SIZE c = (OPPONENT_SIZE)((float)rand() / (float)RAND_MAX * (2.0*(float)count)) ;  
  for(n=0; n<c; n++) {
    OPPONENT_SIZE t ;
    const OPPONENT_SIZE a = (OPPONENT_SIZE)(rand() % count) ;
    const OPPONENT_SIZE b = (OPPONENT_SIZE)(rand() % count) ;
    if (a!=b) {
      t = array[a] ;
      array[a] = array[b] ;
      array[b] = t ;
    }
  }
}
